import { Prisma } from '../db/index.js';
import { File } from '../util/index.js';
class UpdaterVersionService {
    async create(params) {
        const { version } = params;
        return Prisma.updaterVersion.create({ data: { version } });
    }
    async list(options = {}) {
        const { onlyApprovedVersion = false, } = options;
        const where = {
            where: {
                approvedVersion: 'S',
            },
        };
        if (onlyApprovedVersion) {
            const result = await Prisma.updaterVersion.findFirst(where);
            const { updatedAt: published, version } = result || {};
            return version ? { version, published } : {};
        }
        return Prisma.updaterVersion.findMany();
    }
    async download({ version }, response) {
        let fileName = '';
        const handleDownloadError = (error) => {
            console.log(error);
            try {
                response.status(500).send({
                    error: `Ocorreu um erro ao realizar download da versão ${fileName}`,
                    message: error.message,
                    stack: error.stack,
                });
            }
            catch (err) {
                console.error(err);
            }
        };
        try {
            const { UPDATER_VERSIONS_PATH: filePath } = process.env;
            let versionToDownload = version;
            if (version) {
                fileName = `${version}.tar.gz`;
            }
            else {
                const { version: approvedVersion = null } = await this.list({ onlyApprovedVersion: true });
                if (approvedVersion) {
                    versionToDownload = approvedVersion;
                    fileName = `${approvedVersion}.tar.gz`;
                }
                else {
                    return response.status(400).send({ error: 'Não existe versão aprovada e não foi informada uma versão por parâmetro' });
                }
            }
            const file = new File(fileName, filePath);
            const pathToDownload = file.getFullPath();
            const exists = await file.exists();
            if (!exists)
                return response.status(404).send({ error: `Arquivo ${fileName} não existe` });
            await this.modify({ version: versionToDownload, downloads: { increment: 1 } });
            response.download(pathToDownload, (err) => {
                if (err) {
                    console.error(`Erro ao realizar download da versão ${fileName}`, err);
                    handleDownloadError(err);
                }
            });
        }
        catch (error) {
            handleDownloadError(error);
        }
    }
    async remove({ version, versionsToRemove = [] }) {
        const { UPDATER_VERSIONS_PATH: versionPath, } = process.env;
        const removeVersionFile = async (versionToRemove) => new File(`${versionToRemove}.tar.gz`, versionPath).remove();
        if (version) {
            await removeVersionFile(version);
        }
        else if (Array.isArray(versionsToRemove) && versionsToRemove.length) {
            await Promise.all(versionsToRemove.map(removeVersionFile));
        }
        const versions = version ? [version] : (versionsToRemove || []);
        const where = { version: { in: versions } };
        const response = Prisma.updaterVersion.deleteMany({ where });
        return response;
    }
    async modify(params) {
        const { version, ...dataToUpdate } = params;
        if (dataToUpdate.approvedVersion === 'S') {
            await Prisma.updaterVersion.updateMany({
                where: { approvedVersion: 'S', version: { not: version } },
                data: { approvedVersion: 'N' },
            });
        }
        else if (dataToUpdate.approvedVersion) {
            throw new Error('REQUIRED_ACTIVE_VERSION');
        }
        const response = await Prisma.updaterVersion.update({
            where: { version },
            data: { ...dataToUpdate },
        });
        return response;
    }
}
export default new UpdaterVersionService();
