import stream from 'stream';
import DockerRode from 'dockerode';
import { Assistant, Oracle } from '../../config/index.js';
import { promisify } from '../../util/index.js';
import { Prisma } from '../../db/index.js';
import { DockerPrettyInfo, SocketService } from '../index.js';
class Docker {
    constructor() {
        Docker.allInstances = {};
        Prisma.server.findMany().then((servers) => {
            servers.forEach((server) => {
                const { name, port, protocol, host, deployEnabled } = server;
                const handledProtocol = protocol === 'https' ? 'https' : 'http';
                this.addInstance(name, deployEnabled, {
                    port,
                    protocol: handledProtocol,
                    host,
                });
            }, {});
        });
    }
    connect(params) {
        params.port = params.port || '2375';
        return new DockerRode(params);
    }
    addInstance(name, deployEnabled, params) {
        Docker.allInstances[name] = this.connect(params);
        Docker.allInstances[name].serverName = name;
        Docker.allInstances[name].host = params.host;
        Docker.allInstances[name].deployEnabled = deployEnabled;
        return Docker.allInstances[name];
    }
    getInstances() {
        return Docker.allInstances;
    }
    getInstanceByHost(host) {
        const allInstances = Object.values(Docker.allInstances);
        const serverByHost = allInstances.find((server) => server.host === host);
        return serverByHost || null;
    }
    removeInstance(name) {
        delete Docker.allInstances[name];
    }
    removeInstances(names) {
        names.forEach(this.removeInstance);
    }
    getInstance({ serverName }) {
        if (!serverName)
            throw new Error('Not found {serverName}');
        if (serverName === 'all')
            return Object.values(Docker.allInstances);
        const instance = Docker.allInstances[serverName];
        if (serverName.includes(',')) {
            const servers = serverName.split(',');
            const filteredInstances = Object.values(Docker.allInstances).filter((curInstance) => servers.includes(curInstance.serverName));
            return filteredInstances;
        }
        if (!instance)
            throw new Error('Invalid {serverName} not found instance');
        return instance;
    }
    async start({ instance, containerName }) {
        const container = instance.getContainer(containerName);
        try {
            await promisify(container.start, container);
            return { status: 'started', id: container.id };
        }
        catch (error) {
            return Promise.reject(error);
        }
    }
    setPortBinds(config, port) {
        config.ExposedPorts = {};
        const tcpPort = `${port}/tcp`;
        config.ExposedPorts[tcpPort] = {};
        config.HostConfig = config.HostConfig || {};
        config.HostConfig.PortBindings = config.HostConfig.PortBindings || {};
        config.HostConfig.PortBindings[tcpPort] = [{ HostPort: port.toString() }];
    }
    async create(params) {
        const { serverName, containerName, orgId, mode = 'sync' } = params;
        if (!orgId || !containerName || !serverName) {
            throw new Error('Invalid Params {containerName} or {orgId} or {serverName}');
        }
        try {
            let config = Oracle.getConfig({
                name: containerName,
            });
            config = Assistant.getConfig(config, params);
            this.setPortBinds(config, orgId);
            return await new Promise((resolve, reject) => {
                if (mode !== 'sync')
                    resolve({ status: 'in_progress' });
                const instance = this.getInstance({ serverName });
                instance
                    .createContainer(config)
                    .then(async (container) => {
                    let errorInfo = null;
                    await this.start({ instance, containerName }).catch((error) => {
                        errorInfo = error;
                    });
                    const responseError = { status: 'fail', error: errorInfo };
                    const options = {
                        filters: { id: [container.id] },
                    };
                    const responseSuccess = await this.list({ instance, options });
                    const response = errorInfo
                        ? responseError
                        : DockerPrettyInfo.prettyStats(responseSuccess.containers[0]);
                    if (mode === 'sync') {
                        if (errorInfo) {
                            reject(response);
                        }
                        else {
                            resolve(Object.assign(response, {
                                serverName,
                                name: containerName,
                                id: container.id,
                            }));
                        }
                    }
                    else {
                        SocketService.emit({ page: 'container-management', event: 'create' }, Object.assign(response, {
                            serverName,
                            name: containerName,
                            id: container.id,
                        }));
                    }
                })
                    .catch((error) => {
                    console.error('Failed to create container\n', JSON.stringify(config, null, 2));
                    reject(error);
                });
            });
        }
        catch (err) {
            return Promise.reject(err);
        }
    }
    async restart({ instance, containerName }) {
        try {
            const container = instance.getContainer(containerName);
            const result = await promisify(container.restart, container);
            return { status: 'restarted', result };
        }
        catch (error) {
            return { error };
        }
    }
    async modify({ serverName, containerName, type }) {
        const modifyFunctions = {
            stop: this.stop,
            start: this.start,
            restart: this.restart,
            default() {
                throw new Error('NOT_FOUND_RESOURCE');
            },
        };
        const functionToCall = modifyFunctions[type || 'default'] || modifyFunctions.default;
        return functionToCall.apply(this, [
            { instance: this.getInstance({ serverName }), containerName },
        ]);
    }
    async stop({ instance, containerName }) {
        const container = instance.getContainer(containerName);
        try {
            await promisify(container.stop, container);
            return { status: 'stopped', id: container.id };
        }
        catch (error) {
            return Promise.reject(error);
        }
    }
    async removeMultipleContainers(containersToRemove) {
        const promises = [];
        containersToRemove.forEach(({ containerName, serverName, shortId }) => {
            promises.push(this.removeSingleContainer({
                containerName,
                serverName,
                shortId,
            }));
        });
        const result = await Promise.allSettled(promises);
        const response = {
            success: [],
            error: [],
        };
        result.forEach((res) => {
            if (res.status === 'fulfilled') {
                response.success.push(res.value);
            }
            else {
                response.error.push(res.reason);
            }
        });
        return response;
    }
    async removeSingleContainer({ serverName, containerName, shortId }) {
        try {
            if (!serverName || !containerName) {
                throw new Error('');
            }
            const instance = this.getInstance({ serverName });
            const container = await instance.getContainer(containerName);
            await this.stop({ instance, containerName }).catch((err) => {
                if (err.reason !== 'container already stopped') {
                    throw new Error(err.reason);
                }
            });
            await promisify(container.remove, container);
            return { status: 'removed', containerName, shortId };
        }
        catch (error) {
            console.error(error);
            return Promise.reject(error);
        }
    }
    async remove(params) {
        try {
            const { containerName, serverName, containersToRemove } = params;
            if (containerName) {
                return await this.removeSingleContainer({
                    serverName,
                    containerName,
                });
            }
            if (containersToRemove) {
                return await this.removeMultipleContainers(containersToRemove);
            }
        }
        catch (error) {
            console.error(error);
            return Promise.reject(error);
        }
    }
    async list({ instance, options = { all: false } }) {
        try {
            const instanceCollection = Array.isArray(instance)
                ? instance
                : [instance];
            const allPromise = [];
            instanceCollection.forEach((curInstance) => {
                allPromise.push(promisify(curInstance.listContainers, curInstance, options, {
                    serverName: curInstance.serverName,
                }));
            });
            const response = await Promise.allSettled(allPromise);
            return response.reduce((result, info) => {
                if (info.status === 'rejected') {
                    result.errors = result.errors.concat(info.reason);
                }
                else {
                    result.containers = result.containers.concat(info.value);
                }
                return result;
            }, { errors: [], containers: [] });
        }
        catch (error) {
            return Promise.reject(error);
        }
    }
    async listAll({ instance }) {
        const options = { all: true };
        return this.list({ instance, options });
    }
    async getContainersPerServer({ instance }) {
        const { errors, containers } = await this.listAll({ instance });
        const serversMap = {};
        if (Array.isArray(instance) && instance.length) {
            instance.forEach((server) => {
                serversMap[server.serverName] = {
                    serverName: server.serverName,
                    containersCount: 0,
                    containers: [],
                    deployEnabled: server.deployEnabled,
                };
            });
        }
        const containersPerServer = containers.reduce((serverMap, containerInfo) => {
            const { serverName, ...container } = containerInfo;
            serverMap[serverName] = serverMap[serverName] || { containersCount: 0, containers: [], serverName };
            const server = serverMap[serverName];
            server.containersCount += 1;
            server.containers.push(container);
            server.deployEnabled = Docker.allInstances[serverName].deployEnabled;
            return serverMap;
        }, serversMap);
        return {
            errors,
            servers: containersPerServer,
        };
    }
    async summary({ instance }) {
        const { errors, containers } = await this.listAll({ instance });
        const summary = {};
        const promiseStats = [];
        containers.forEach((info) => {
            const containerInfo = {};
            const [name] = info.Names;
            const { Id: id, serverName } = info;
            Object.assign(containerInfo, {
                id: id.substring(0, 12),
                name: name.replace('/', ''),
                serverName,
            });
            summary[id] = containerInfo;
            const container = this.getInstance({ serverName }).getContainer(id);
            promiseStats.push(promisify(container.stats, container, { stream: false }).then((stats) => {
                try {
                    Object.assign(summary[id], DockerPrettyInfo.prettyStats(stats), DockerPrettyInfo.prettyContainerInfo(info));
                }
                catch (error) {
                    console.error(error);
                }
            }));
        });
        await Promise.allSettled(promiseStats);
        return { containers: Object.values(summary), errors };
    }
    createLogStream(serverName, containerName, stdname) {
        const logId = `${serverName}_${containerName}_${stdname}`;
        Docker.logStream[logId] = Docker.logStream[logId] || new stream.PassThrough();
        return Docker.logStream[logId];
    }
    getLogStreamId(serverName, containerName, stdname) {
        const logId = `${serverName}_${containerName}_${stdname}`;
        return logId;
    }
    exitsLogStream(serverName, containerName, stdname) {
        return !!Docker.logStream[this.getLogStreamId(serverName, containerName, stdname)];
    }
    async log({ instance, containerName, options }) {
        try {
            if (!containerName)
                throw new Error('Invalid param {{containerName}}');
            const container = instance.getContainer(containerName);
            const { serverName } = instance;
            if (options.mode === 'async') {
                if (this.exitsLogStream(serverName, containerName, 'stream'))
                    return { message: 'log watch already created' };
                let logStream = this.createLogStream(serverName, containerName, 'stream');
                logStream.on('data', (chunk) => {
                    SocketService.emit({ page: 'container-management', event: 'logs' }, {
                        serverName,
                        containerName,
                        log: chunk.toString('utf8'),
                    });
                });
                const fiveMinutesAgo = new Date(new Date().getTime() - (5 * 60 * 1000));
                container.logs({
                    follow: true,
                    stdout: true,
                    stderr: true,
                    since: Math.floor(fiveMinutesAgo.getTime() / 1000),
                }, (err, dockerStream) => {
                    if (err)
                        return console.error(err);
                    container.modem.demuxStream(dockerStream, logStream, logStream);
                    dockerStream.on('end', () => {
                        logStream.end('!stop!');
                        logStream.destroy();
                        Docker.logStream[this.getLogStreamId(serverName, containerName, 'stream')] = null;
                        logStream = null;
                    });
                });
                return { message: 'log-created' };
            }
            const logs = await promisify(container.logs, container, {
                tail: 50,
                stdout: true,
                stderr: true,
                timestamps: true,
            });
            const logString = Buffer.from(logs).toString('utf8');
            return logString.split('\n').map((line) => {
                const result = line.match(/\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+Z/i) || [];
                const [timestamp] = result;
                const timestampIdx = result.index || 0;
                const dateInstance = new Date(timestamp);
                const dateStr = dateInstance.toLocaleDateString('pt-BR');
                const timeStr = dateInstance.toLocaleTimeString('pt-BR');
                const formattedTimeStamp = dateStr === 'Invalid Date' ? timestamp : `${dateStr} ${timeStr}`;
                return line.substring(timestampIdx).replace(timestamp, formattedTimeStamp);
            }).join('\n');
        }
        catch (error) {
            return Promise.reject(error);
        }
    }
    async inspect({ instance, containerName }) {
        const container = instance.getContainer(containerName);
        try {
            const info = await promisify(container.inspect, container);
            return info;
        }
        catch (error) {
            return Promise.reject(error);
        }
    }
    async servers() {
        return { servers: Object.keys(Docker.allInstances) };
    }
    async info(params) {
        const { serverName = 'all', containerName, type, ...options } = params;
        const infoFunctions = {
            list: this.list,
            log: this.log,
            inspect: this.inspect,
            summary: this.summary,
            listAll: this.listAll,
            servers: this.servers,
            default() {
                throw new Error('NOT_FOUND_RESOURCE');
            },
        };
        if (type
            && containerName === 'all'
            && ['list', 'listAll', 'summary'].includes(type)) {
            throw new Error('Not found {containerName} parameter');
        }
        const functionToCall = type && infoFunctions[type] ? infoFunctions[type] : infoFunctions.default;
        return functionToCall.apply(this, [
            { instance: this.getInstance({ serverName: serverName || 'all' }), containerName, options },
        ]);
    }
    async statusContainer({ containerName, host, protocol = 'http', port }) {
        const connectionParams = { host, protocol, port };
        const status = {
            exists: false,
            running: false,
            error: false,
            serverName: '',
        };
        if (!host)
            throw new Error('Invalid param {host}');
        if (!containerName)
            throw new Error('Invalid param {containerName}');
        const server = this.getInstanceByHost(host) || this.connect(connectionParams);
        status.serverName = server.serverName || '';
        const options = {
            filters: { name: [containerName] },
            all: true,
        };
        const { errors, containers } = await this.list({ instance: server, options });
        if (containers.length) {
            const [container] = containers;
            status.exists = true;
            status.running = container.State === 'running';
        }
        else if (errors.length) {
            const [error] = errors;
            status.error = error;
        }
        return status;
    }
}
Docker.allInstances = {};
Docker.logStream = {};
export default new Docker();
